from typing import Text

import pytest

from alembic.operations import Operations
from alembic.runtime.migration import MigrationContext
import sqlalchemy as sa
from sqlalchemy.engine import RowProxy
from sqlalchemy.orm import Session

from rasax.community.database.schema_migrations.alembic import utils as migration_utils


@pytest.fixture
def alembic_op(session: Session):
    migration_context = MigrationContext.configure(session.get_bind())

    op = Operations(migration_context)

    # Replace `op` import for the time of the tests.
    old_op = migration_utils.op
    migration_utils.op = op

    yield op

    # Revert replacing of the `op` import
    migration_utils.op = old_op


def test_modify_columns(alembic_op: Operations, session: Session):
    table_name = "test_table"
    test_bool_column = "test_bool_column"
    test_string_column = "test_string_column"

    # Create test table
    alembic_op.create_table(
        table_name,
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(test_bool_column, sa.Boolean, nullable=False),
        sa.Column(test_string_column, sa.String, nullable=False),
    )

    metadata = sa.MetaData(bind=alembic_op.get_bind())
    metadata.reflect()
    table = sa.Table(table_name, metadata, autoload=True)

    # Insert fake data
    session.execute(table.insert().values((1, False, "some text")))
    session.execute(table.insert().values((2, True, "some other text")))
    session.commit()

    # Assert fake data is there
    entries = session.execute(table.select()).fetchall()
    assert entries == [(1, False, "some text"), (2, True, "some other text")]

    # Transform columns and values in column
    test_bool_transformation = migration_utils.ColumnTransformation(
        test_bool_column,
        new_column_kwargs={"type_": sa.String()},
        modify_from_column_value=lambda x: str(x) + "_test",
    )
    test_string_transformation = migration_utils.ColumnTransformation(
        test_string_column,
        new_column_args=[sa.Text()],
        new_column_kwargs={"nullable": True},
    )

    migration_utils.modify_columns(
        table_name, [test_bool_transformation, test_string_transformation]
    )

    # Update table reflection
    metadata.reflect(extend_existing=True)
    table = sa.Table(table_name, metadata, autoload=True)

    # Test if new value is expected
    entries = session.execute(table.select()).fetchall()

    assert entries == [
        (1, "False_test", "some text"),
        (2, "True_test", "some other text"),
    ]


def test_modify_columns_from_row_with_copy(alembic_op: Operations, session: Session):
    table_name = "test_modify_columns_from_row_with_copy"
    test_bool_column = "test_bool_column"
    test_string_column = "test_string_column"

    # Create test table
    alembic_op.create_table(
        table_name,
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(test_bool_column, sa.Boolean, nullable=False),
        sa.Column(test_string_column, sa.String, nullable=False),
    )
    metadata = sa.MetaData(bind=alembic_op.get_bind())
    metadata.reflect()
    table = sa.Table(table_name, metadata, autoload=True)

    # Insert fake data
    session.execute(table.insert().values((1, False, "some text")))
    session.execute(table.insert().values((2, True, "some other text")))
    session.commit()

    def modify_from_row(row: RowProxy) -> Text:
        return f"bool: {row.test_bool_column}, text: {row.test_string_column}"

    # Transform columns and values in column
    new_column_name = "text_bool_combined"
    test_bool_transformation = migration_utils.ColumnTransformation(
        test_bool_column,
        new_column_kwargs={"type_": sa.String()},
        modify_from_row=modify_from_row,
        new_column_name=new_column_name,
    )
    migration_utils.modify_columns(table_name, [test_bool_transformation])

    # Update table reflection
    metadata.reflect(extend_existing=True)
    table = sa.Table(table_name, metadata, autoload=True)

    # Test if new value is expected
    entries = session.execute(table.select()).fetchall()

    assert entries == [
        (1, False, "some text", "bool: False, text: some text"),
        (2, True, "some other text", "bool: True, text: some other text"),
    ]


def test_existing_table_exists(alembic_op: Operations, session: Session):
    table_name = "existing_table"
    test_column = "test_column"

    # Create test table
    alembic_op.create_table(
        table_name,
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(test_column, sa.String, nullable=False),
    )

    assert migration_utils.table_exists(table_name)


def test_nonexistent_table_doesnt_exist(alembic_op: Operations, session: Session):
    assert not migration_utils.table_exists("nonexistent_table")


def test_table_has_existing_column(alembic_op: Operations, session: Session):
    table_name = "test_column_table"
    test_column = "existing_column"

    # Create test table
    alembic_op.create_table(
        table_name,
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(test_column, sa.String, nullable=False),
    )

    assert migration_utils.table_has_column(table_name, test_column)


def test_table_doesnt_have_nonexistent_column(alembic_op: Operations, session: Session):
    table_name = "test_column_table_2"
    test_column = "existing_column"

    # Create test table
    alembic_op.create_table(
        table_name,
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(test_column, sa.String, nullable=False),
    )

    assert not migration_utils.table_has_column(table_name, "nonexistent_column")


def test_set_synonym_references_for_old_mapped_values(
    alembic_op: Operations, session: Session
):
    metadata = sa.MetaData(bind=alembic_op.get_bind())
    metadata.reflect()

    # Insert entity
    entity_table = sa.Table("nlu_training_data_entity", metadata, autoload=True)
    session.execute(entity_table.insert().values(original_value="hut"))

    # Insert synonym
    synonym_table = sa.Table("entity_synonym", metadata, autoload=True)
    session.execute(
        synonym_table.insert().values(id=1, name="house", project_id="default")
    )

    # Insert mapping
    synonym_value_table = sa.Table("entity_synonym_value", metadata, autoload=True)
    session.execute(
        synonym_value_table.insert().values(entity_synonym_id=1, name="hut")
    )
    session.commit()

    # Run migration function
    from rasax.community.database.schema_migrations.alembic.versions import (
        migration_2019_10_25_add_column_for_synonym_foreign_key_a813c6c15a4c as migration,
    )

    migration.op = alembic_op
    migration._set_synonym_references_for_old_mapped_values()

    # Check if value was updated
    rows = session.execute(entity_table.select()).fetchall()
    assert rows[0]["entity_synonym_id"] == 1
